/* Management information forwarding between AN <-> LE */

/* (C) 2022 by Andreas Eversberg <jolly@eversberg.eu>
 *
 * All Rights Reserved
 *
 * SPDX-License-Identifier: GPL-2.0+
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <stdint.h>
#include <osmocom/gsm/tlv.h>

#include "v5x_internal.h"
#include "v5x_protocol.h"
#include "v5x_le_port_fsm.h"
#include "v5x_le_pstn_fsm.h"
#include "logging.h"

void v5x_le_mgmt_init(void)
{
	LOGP(DV5MGMT, LOGL_NOTICE, "Using V5x management\n");
}

/* received management primitives from LE (NAT protocol via PH-socket) */
void v5x_mph_rcv_le(struct v5x_user_port *v5up, uint8_t prim, uint8_t *data, int length)
{
	ph_socket_t *s = &v5up->ph_socket;

	switch (prim) {
	case PH_PRIM_CTRL_REQ:
		if (length && *data == PH_CTRL_UNBLOCK) {
			LOGP(DV5MGMT, LOGL_INFO, "Received unblocking from LE of port %d.\n", v5up->nr);
			v5up->le_unblocked = true;
			v5x_mph_snd(v5up, MPH_UBR);
			if (v5up->pstn.proto && v5up->an_unblocked)
				v5x_le_pstn_mdu_snd(v5up, MDU_CTRL_port_unblocked);
			if (!v5up->an_unblocked) {
				LOGP(DV5MGMT, LOGL_INFO, "Reply with blocking, because AN is blocked.\n");
				uint8_t ctrl = PH_CTRL_BLOCK;
				ph_socket_tx_msg(&v5up->ph_socket, v5up->type == V5X_USER_TYPE_ISDN ? 3 : 0, PH_PRIM_CTRL_IND, &ctrl, 1);
			}
		}
		if (length && *data == PH_CTRL_BLOCK) {
			LOGP(DV5MGMT, LOGL_INFO, "Received blocking from LE of port %d.\n", v5up->nr);
			v5up->le_unblocked = false;
			v5x_mph_snd(v5up, MPH_BI);
			if (v5up->pstn.proto)
				v5x_le_pstn_mdu_snd(v5up, MDU_CTRL_port_blocked);
		}
		break;
	case PH_PRIM_ACT_REQ:
		LOGP(DV5MGMT, LOGL_INFO, "Received activation from LE of port %d.\n", v5up->nr);
		if (!v5up->an_unblocked) {
			ph_socket_tx_msg(s, 3, PH_PRIM_DACT_IND, NULL, 0);
			break;
		}
		v5x_mph_snd(v5up, MPH_AR);
		break;
	case PH_PRIM_DACT_REQ:
		LOGP(DV5MGMT, LOGL_INFO, "Received deactivation from LE of port %d.\n", v5up->nr);
		v5x_mph_snd(v5up, MPH_DR);
		break;
	default:
		LOGP(DV5MGMT, LOGL_INFO, "Received unknown prim %d from LE for port %d.\n", prim, v5up->nr);
	}
}

static void _block_unblock_socket(struct v5x_user_port *v5up, bool block)
{
	uint8_t ctrl;

	if (block && v5up->an_unblocked) {
		v5up->an_unblocked = 0;
		ctrl = PH_CTRL_BLOCK;
		ph_socket_tx_msg(&v5up->ph_socket, v5up->type == V5X_USER_TYPE_ISDN ? 3 : 0, PH_PRIM_CTRL_IND, &ctrl, 1);
		return;
	}
	if (!block && !v5up->an_unblocked) {
		v5up->an_unblocked = 1;
		ctrl = PH_CTRL_UNBLOCK;
		ph_socket_tx_msg(&v5up->ph_socket, v5up->type == V5X_USER_TYPE_ISDN ? 3 : 0, PH_PRIM_CTRL_IND, &ctrl, 1);
		return;
	}
}

/* received management primitives from AN (via port control protocol) */
void v5x_mph_rcv_an(struct v5x_user_port *v5up, enum v5x_mph_prim prim, uint8_t perf_grading)
{
	switch (prim) {
	case MPH_UBR:
		LOGP(DV5MGMT, LOGL_INFO, "Received unblocking request from AN of port %d.\n", v5up->nr);
		if (v5up->le_unblocked) {
			LOGP(DV5MGMT, LOGL_INFO, "Reply with unblocking request.\n");
			v5x_mph_snd(v5up, MPH_UBR);
			_block_unblock_socket(v5up, false);
			if (v5up->pstn.proto)
				v5x_le_pstn_mdu_snd(v5up, MDU_CTRL_port_unblocked);
		} else
			LOGP(DV5MGMT, LOGL_INFO, "Ignore unblocking request from AN, because LE is blocked.\n");
		break;
	case MPH_UBI:
		LOGP(DV5MGMT, LOGL_INFO, "Received unblocking indication from AN of port %d.\n", v5up->nr);
		_block_unblock_socket(v5up, false);
		if (v5up->pstn.proto)
			v5x_le_pstn_mdu_snd(v5up, MDU_CTRL_port_unblocked);
		break;
	case MPH_BR:
		LOGP(DV5MGMT, LOGL_INFO, "Received blocking request from AN of port %d.\n", v5up->nr);
		v5x_mph_snd(v5up, MPH_BI);
		_block_unblock_socket(v5up, true);
		if (v5up->pstn.proto)
			v5x_le_pstn_mdu_snd(v5up, MDU_CTRL_port_blocked);
		break;
	case MPH_BI:
		LOGP(DV5MGMT, LOGL_INFO, "Received blocking indication from AN of port %d.\n", v5up->nr);
		_block_unblock_socket(v5up, true);
		if (v5up->pstn.proto)
			v5x_le_pstn_mdu_snd(v5up, MDU_CTRL_port_blocked);
		break;
	case MPH_AI:
		LOGP(DV5MGMT, LOGL_INFO, "Received activation indication from AN of port %d.\n", v5up->nr);
		ph_socket_tx_msg(&v5up->ph_socket, 3, PH_PRIM_ACT_IND, NULL, 0);
		break;
	case MPH_DI:
		LOGP(DV5MGMT, LOGL_INFO, "Received deactivation indication from AN of port %d.\n", v5up->nr);
		ph_socket_tx_msg(&v5up->ph_socket, 3, PH_PRIM_DACT_IND, NULL, 0);
		break;
	default:
		LOGP(DV5MGMT, LOGL_INFO, "Received unknown prim %d from AN for port %d.\n", prim, v5up->nr);
	}
}

/* receive FE-message from NAT (PH-socket) and forward it to the PSTN protocol */
void v5x_nat_rcv_fe(struct v5x_user_port *v5up, struct msgb *msg)
{
	enum v5x_fe_prim prim;

	prim = *msg->data;
	msgb_pull(msg, 1);

	LOGP(DV5MGMT, LOGL_INFO, "FE message: NAT -> PSTN_LE of port %d.\n", v5up->nr);
	v5x_le_pstn_fe_snd(v5up, prim, msg);
}

/* receive FE-message from PSTN protocol and forward it to NAT (PH-socket) */
void v5x_pstn_rcv_fe(struct v5x_user_port *v5up, enum v5x_fe_prim prim, struct msgb *msg)
{
	*msgb_push(msg, 1) = prim;

	LOGP(DV5MGMT, LOGL_INFO, "FE message: PSTN_LE -> NAT of port %d.\n", v5up->nr);
	ph_socket_tx_msg(&v5up->ph_socket, 0, PH_PRIM_DATA_IND, msg->data, msg->len);

	msgb_free(msg);
}

/* receive MDU primitives from PSTN protocol */
void v5x_pstn_rcv_mdu(struct v5x_user_port *v5up, enum v5x_mgmt_prim prim)
{
	switch (prim) {
	case MDU_error_ind:
		LOGP(DV5MGMT, LOGL_NOTICE, "Received MDU-error_indication from PSTN protocol of port %d.\n", v5up->nr);
		break;
	case MDU_CTRL_port_restart_ack:
		//FIXME: handle restart procedure
		break;
	default:
		LOGP(DV5MGMT, LOGL_INFO, "Received unknown prim %d from PSTN protocol for port %d.\n", prim, v5up->nr);
	}
}
